/* 
   cadaver, command-line DAV client
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
   $Id: cadaver.c,v 1.45 1999/12/22 23:55:34 joe Exp $
*/

#include <config.h>

#include <sys/types.h>

#include <stdio.h>
#include <ctype.h>

#include <signal.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif 
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

/* joe: My man pages say getpass() should be in pwd.h...
 * looks like getpass() is a BSD-ism, what is the POSIX equiv? */
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#ifdef HAVE_READLINE_H
#include <readline.h>
#elif HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif

#ifdef HAVE_HISTORY_H
#include <history.h>
#elif HAVE_READLINE_HISTORY_H
#include <readline/history.h>
#endif

#include <errno.h>

#ifndef HAVE_SNPRINTF
#include "snprintf.h"
#endif

#include "getopt.h"

#ifdef ENABLE_NETRC
#include "netrc.h"
#endif

#include "frontend.h"
#include "protocol.h"
#include "httpdav.h"
#include "string_utils.h"
#include "common.h"

#include "cadaver.h"
#include "cmdline.h"
#include "commands.h"
#include "options.h"

#ifdef ENABLE_NETRC
static netrc_entry *netrc_list; /* list of netrc entries */
#endif

static struct proto_host server = {0}; /* server details */
char *path, /* current working collection */
    *old_path; /* previous working collection */
/* TODO: use dynamically allocated memory for paths, maybe */

static char *progname; /* argv[0] */

int tolerant; /* tolerate DAV-enabledness failure */

bool have_connection; /* true when we have server details */

/* Protoypes */

static void usage(void);
static void parse_args( int argc, char *argv[], struct proto_host *server );
static char *read_command(void);

static void quit_handler( int signal );

static void close_connection( void );
static void open_connection( const char *hostname, const char *path );

static void init_signals( void );
static void init_netrc( void );

/* Real globals */
char transfer_msg[BUFSIZ];

enum transfer_status transfer_status;

#define CMD_VARY 9999

/* Separate structures for commands and command names. */
const struct command commands[] = {
/* C1: connected, 1-arg function C2: connected, 2-arg function
 * U0: disconnected, 0-arg function. */
#define C1(x,c,h) { cmd_##x, true, 1, 1, globscope_remote, execute_##x,c,h }
#define U0(x,h) { cmd_##x, false, 0, 0, globscope_none, execute_##x,#x,h }
#define UO1(x,c,h) { cmd_##x, false, 0, 1, globscope_none, execute_##x,c,h }
#define C2M(x,c,h) { cmd_##x, true, 2, CMD_VARY, globscope_remote, multi_##x,c,h }
#define C1M(x,c,h) { cmd_##x, true, 1, CMD_VARY, globscope_remote, multi_##x,c,h }
    { cmd_ls, true, 0, 1, globscope_none, execute_ls, 
      "ls [path]", "List contents of current [or other] collection" },
    C1(cd, "cd path", "Change to specified collection"),
    { cmd_put, true, 1, 2, globscope_none, execute_put, 
      "put local [remote]", "Upload local file" },
    { cmd_get, true, 1, 2, globscope_none,
      execute_get, "get remote [local]", "Download remote resource" },
    C1M( mget, "mget remote...", "Download many remote resources" ),
    { cmd_mput, true, 1, CMD_VARY, globscope_local, multi_mput, 
      "mput local...", "Upload many local files" },
    C1M(less, "less remote...", "Display remote resource through pager"), 
    C1M(mkcol, "mkcol remote...", "Create remote collection(s)"), 
    C1M(cat, "cat remote...", "Display remote resource(s)"), 
    C1M(delete, "delete remote...", "Delete remote resource(s)"),
    C2M(copy, "copy source dest", "Copy resource from source to dest"), 
    C2M(move, "move source... dest", "Move resource(s) from source to dest"),
    { cmd_set, false, 0, 2, globscope_none, execute_set, 
      "set [option] [value]", "Set an option, or display options" }, 
    { cmd_open, false, 1, 2, globscope_none, open_connection, 
      "open host[:port] [/path/]","Open connection to server at path" },
    { cmd_close, true, 0, 0, globscope_none,
      close_connection, "close", "Close current connection" },
    { cmd_echo, false, 1, CMD_VARY, globscope_remote,
      execute_echo, "echo", NULL },
    { cmd_quit, false, 0, 1, globscope_none, NULL, "quit", "Exit program" },
    /* Unconnected operation, 1 mandatory argument */
    { cmd_unset, false, 1, 2, globscope_none, execute_unset, 
      "unset [option] [value]", "Unsets or clears value from option." },
    /* Unconnected operation, 0 arguments */
    UO1(help, "help [command]", "Display help message" ), 
    UO1(lcd, "lcd [directory]", "Change local working directory"), 
    U0(lpwd, "Print local working directory" ),
    { cmd_lls, false, 0, CMD_VARY, globscope_local, execute_lls, 
      "lls [options]", "Display local directory listing" },
    { cmd_unknown, 0 } /* end-of-list marker, DO NOT move */
#undef C1
#undef U0
#undef UO1
#undef C2M
#undef C1M
};    

static void usage() {
    printf( 
"Usage: %s [OPTIONS] [hostname[:port] [path]]\n"
"  Port defaults to 80, path defaults to '/'\n"
"Options:\n"
"  -e, --expect100  Enable sending of `Expect: 100-continue' header.\n"
"     *** Warning: For Apache servers, use only with version 1.3.9 and later.\n"
"  -t, --tolerant   Allow cd/open into non-DAV enabled collection.\n"
"  -V, --version    Display version information.\n"
"  -h, --help       Display this help message.\n"
"Please send bug reports and feature requests to <cadaver@webdav.org>\n", progname );
}

static void version( void ) {
    printf( PACKAGE " " VERSION "\n" );
}

static void close_connection( void ) {
    http_finish( );
    have_connection = false;
    if( path ) free( path );
    if( old_path ) free( old_path );
    path = NULL;
    printf( "Connection closed.\n" );
}

static void open_connection( const char *hostname, const char *open_path ) {
    char *pnt, *real_hostname = strdup(hostname);
    const char *use_path;
    if( have_connection ) {
	close_connection( );
    } else {
	if( path ) free( path );
	if( old_path ) free( old_path );
    }
    pnt = strchr( real_hostname, ':' );
    if( pnt != NULL && (strlen(pnt+1)>0) ) {
	*pnt = '\0';
	server.port = atoi( pnt+1 );
    } else {
	server.port = 80;
    }
    if( server.hostname ) free( server.hostname );
    server.hostname = real_hostname;
#ifdef ENABLE_NETRC
    {
	netrc_entry *found;
	found = search_netrc( netrc_list, hostname );
	if( found != NULL ) {
	    if( found->account && found->password ) {
		server.username = found->account;
		server.password = found->account;
	    }
	}
    }
#endif /* ENABLE_NETRC */
    have_connection = false;
    /* Default to '/' if no path given */
    use_path = open_path?open_path:"/";
    http_init_checks = false; /* we're going to do a PROPFIND, so no worries */
    switch( http_init( use_path, &server, NULL ) ) {
    case PROTO_OK:
	have_connection = true;
	path = NULL;
	execute_cd( use_path );
	if( !path ) {
	    close_connection( );
	}
	break;
    case PROTO_LOOKUP:
	printf( "Could not resolve hostname.\n" );
	break;
    case PROTO_CONNECT:
	printf( "Could not connect to remote host.\n" );
	break;
    default:
	printf( "Could not open connection:\n%s\n", http_error );
	break;
    }
}
	    
static void parse_args( int argc, char *argv[], 
		       struct proto_host *server ) {
    static const struct option opts[] = {
	{ "version", no_argument, NULL, 'V' },
	{ "help", no_argument, NULL, 'h' },
	{ "expect100", no_argument, NULL, 'e' },
	{ "tolerant", no_argument, NULL, 't' },
	{ 0, 0, 0, 0 }
    };
    int optc;
    while( (optc = getopt_long( argc, argv, "ehV", opts, NULL )) != -1 ) {
	switch( optc ) {
	case 'h': usage( ); exit( -1 );
	case 'V': version( ); exit( -1 );
	case 'e': http_enable_expect = true; break;
	case 't': tolerant = 1; break;
	case '?': 
	default:
	    printf( "Try `%s --help' for more information.\n", progname );
	    exit(-1);
	}
    }
    if( argc == 1 ) {
	/* could do open to "localhost" here, but...
	 * principle of least surprise. */
	path = NULL;
	have_connection = false;
    } else if( argc == 2 ) {
	open_connection( argv[1], NULL );
#ifdef HAVE_ADD_HISTORY
	{
	    char *cmd;
	    CONCAT2( cmd, "open ", argv[1] );
	    add_history( cmd );
	    free( cmd );
	}
#endif
    } else if( argc == 3 ) {
	open_connection( argv[1], argv[2] );
#ifdef HAVE_ADD_HISTORY
	{
	    char *cmd;
	    CONCAT4( cmd, "open ", argv[1], " ", argv[2] );
	    add_history( cmd );
	    free( cmd );
	}
#endif
    } else {
	usage( );
	exit( -1 );
    }
}

static char * read_command(void) {
    static char prompt[BUFSIZ];
    if( path ) {
	snprintf( prompt, BUFSIZ, "dav:%s> ", path );
    } else {
	sprintf( prompt, "dav:!> " );
    }
    return readline( prompt ); 
}

/* TODO: nice if the parser handled \ etc */
static int execute_command( const char *line ) {
    const struct command *cmd;
    char **tokens;
    int argcount, ret = 0;
    tokens = parse_command( line, &argcount );
    if( argcount == 0 ) {
	free( tokens );
	return 0;
    }
    argcount--;
    cmd = get_command( tokens[0] );
    if( cmd == NULL ) {
	printf( "Unrecognised command. Type 'help' for a list of commands.\n");
    } else if( argcount < cmd->min_args ) {
	printf( "The `%s' command requires %d argument%s:\n  %s : %s\n",
		tokens[0], cmd->min_args, cmd->min_args==1?"":"s",
		cmd->call, cmd->short_help );
    } else if( argcount > cmd->max_args ) {
	if( cmd->max_args ) {
	    printf( "The `%s' command takes at most %d argument%s:\n"
		    "  %s : %s\n", tokens[0], cmd->max_args, 
		    cmd->max_args==1?"":"s", cmd->call, cmd->short_help );
	} else {
	    printf( "The `%s' command takes no arguments:\n  %s : %s\n",
		    tokens[0], cmd->call, cmd->short_help );
	}	    
    } else if( !have_connection && cmd->needs_connection ) {
	printf( "No connection - try `open' first.\n" );
    } else if( cmd->id == cmd_quit ) {
	ret = -1;
    } else {
	/* Cast away */
	void (*take0)(void)=cmd->handler, 
	    (*take1)( const char * )=cmd->handler, 
	    (*take2)( const char *, const char * )=cmd->handler,
	    (*takeV)( int, const char ** )=cmd->handler;
	/* with a nod in the general direction of apache */
	switch( cmd->max_args ) {
	case 0: (*take0)(); break;
	case 1: /* tokens[1]==NULL if argcount==0 */
	    (*take1)( tokens[1] ); break; 
	case 2: 
	    if( argcount <=1 ) {
		(*take2)( tokens[1], NULL );
	    } else {
		(*take2)( tokens[1], tokens[2] );
	    }
	    break;
	case CMD_VARY:
	    (*takeV)( argcount, (const char **) &tokens[1] );
	default:
	    break;
	}
    }
    split_string_free( tokens );
    free( tokens );
    return ret;
}

static void quit_handler( int sig ) {
    /* Reinstall handler */
    if( child_running ) {
	/* The child gets the signal anyway... it can deal with it.
	 * Proper way is probably to ignore signals while child is
	 * running? */
	signal( sig, quit_handler );
	return;
    } else {
	printf( "Terminated by signal %d.\n", sig );
	if( have_connection ) {
	    close_connection( );
	}
	exit( -1 );
    }
}

static void init_signals( void ) {
    signal( SIGPIPE, SIG_IGN );
    signal( SIGTERM, quit_handler );
    signal( SIGABRT, quit_handler );
    signal( SIGQUIT, quit_handler );
    signal( SIGINT, quit_handler );
}

static void init_netrc( void ) {
#ifdef ENABLE_NETRC
    char *netrc;
    CONCAT2( netrc, getenv("HOME"), "/.netrc" );
    netrc_list = parse_netrc( netrc );
#endif
}

int main( int argc, char *argv[] ) {
    int ret = 0;
    progname = argv[0];
    debug_mask = 0;

    init_signals();
    init_netrc();

    parse_args( argc, argv, &server );

    while(ret == 0) {
	char *cmd;
	cmd = read_command();
	if( cmd == NULL ) {
	    /* Is it safe to do this... they just closed stdin, so
	     * is it bad to write to stdout? */
	    putchar( '\n' );
	    ret = 1;
	} else {
#ifdef HAVE_ADD_HISTORY
	    if( strcmp( cmd, "" ) != 0 ) add_history( cmd );
#endif
	    ret = execute_command(cmd);
	    free( cmd );
	}
    }
    if( have_connection ) {
	close_connection( );
    }
    return 0;
}

void fe_connection( fe_conn_status status ) {
    if( transfer_status == tran_going ) return;
    switch( status ) {
    case fe_namelookup:
	printf( "Looking up hostname...\n" );
	break;
    case fe_connecting:
	printf( "Connecting to server...\n" );
	break;
    case fe_connected:
	printf( "Connected.\n" );
	break;
    }	    
}

void fe_transfer_progress( size_t progress, size_t total ) {
    if( transfer_status == tran_none ) return;
    if( transfer_status == tran_starting ) {
	printf( transfer_msg );
    }
    transfer_status = tran_going;
    putchar( '.' );
    fflush( stdout );
}

void fe_login( const char **username, const char **password ) {
    if( transfer_status == tran_going ) {
	printf( "] authentication failed.\n" );
    }
    *username = readline( "Username: " );
    *password = getpass( "Password: " );
    if( transfer_status == tran_going ) {
	printf( "Retrying [" );
	fflush( stdout );
    }
}
